(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nn_db.ml 239 2010-06-23 16:49:03Z gerd $ *)

open Printf
open Plasma_util.Operators

class type transaction =
object
  method connection : Pfs_db.rw_async_connection
  method ro_connection : Pfs_db.ro_async_connection
  method prepare : unit -> unit Uq_engines.engine
  method commit : unit -> unit Uq_engines.engine
  method rollback : unit -> unit Uq_engines.engine
  method event_system : Unixqueue.event_system
  method return : unit -> unit
end

class type ro_transaction =
object
  method connection : Pfs_db.ro_async_connection
  method event_system : Unixqueue.event_system
  method return : unit -> unit
end

type inodeinfo = Pfs_rpcapi_aux.inodeinfo

type modification =
    [ `Datastore_upd of int * string * int64 * bool
    | `Datastore_del of int
    | `Revision_upd of string
    | `Blockalloc_upd of int * int64 * string
    | `Inode_ins of int64 * Pfs_rpcapi_aux.inodeinfo
    | `Inode_upd of int64 * Pfs_rpcapi_aux.inodeinfo
    | `Inode_upd_time of int64 * Pfs_rpcapi_aux.time option * Pfs_rpcapi_aux.time option
    | `Inode_del of int64
    | `Inodeblocks_ins of int64 * Pfs_rpcapi_aux.blockinfo list
    | `Inodeblocks_del of int64 * int64 * int64
    | `Names_ins of string * int64
    | `Names_del of string
   ]

let config = ref (lazy (failwith "Nn_db: module not initialized"))
let db_esys = ref (lazy (failwith "Nn_db: module not initialized"))

let bmaprowsize = Nn_config.bmaprowsize
let bmaprowsizeL = Int64.of_int bmaprowsize

let ro_conns = Queue.create()
  (** inactive ro connections *)

let ro_conns_n = ref 0
  (** #active connectons *)

let ro_conns_e = ref None
  (** This engine exists only if [ro_conns] is empty and no additional
      connection can be created. This engine is done when [ro_conns]
      becomes non-empty.
   *)

let ro_conns_signal = ref (fun () -> ())
  (** This function is called when [ro_conns] becomes non-empty *)


(* TODO: inactive connections should be closed after some period of
   idleness
 *)

let init conf esys =
  config := lazy conf;
  db_esys := lazy esys


let ro_esys() =
  Lazy.force !db_esys


let transact esys =
  let config = Lazy.force !config in
  let c = new Pfs_db.connect config in
  (* The default READ COMMITTED isolation level is just right for us *)
  let _e =
    new Pfs_db.async_exec c esys ~expect:[Postgresql.Command_ok] 
      "START TRANSACTION" in
  let alive = ref true in
  let prepared = ref false in
  let tname =
    "plasma_" ^ (match config#db_dbname with Some n -> n | None -> "") in

  ( object
      method connection = 
	if not !alive then 
	  failwith "Nn_db: connection is already returned to pool";
	c
      method ro_connection = 
	if not !alive then 
	  failwith "Nn_db: connection is already returned to pool";
	(c :> Pfs_db.ro_async_connection)
      method prepare() =
	if not !alive then 
	  failwith "Nn_db: cannot prepare as transaction is already finished";
	let e =
	  new Pfs_db.async_exec c esys ~expect:[Postgresql.Command_ok] 
	    ("PREPARE TRANSACTION '" ^ tname ^ "'") in
	prepared := true;
	Uq_engines.when_state
	  ~is_error:(fun _ -> c # finish)
	  ~is_aborted:(fun _ -> c # finish)
	  e;
	new Uq_engines.map_engine ~map_done:(fun _ -> `Done()) e
      method commit () =
	if not !alive then 
	  failwith "Nn_db: cannot commit as transaction is already finished";
	alive := false;  (* always pretend that from now on *)
	let e =
	  new Pfs_db.async_exec c esys ~expect:[Postgresql.Command_ok] 
	    (if !prepared then
	       "COMMIT PREPARED '" ^ tname ^ "'"
	     else
	       "COMMIT") in
	Uq_engines.when_state
	  ~is_done:(fun _ -> c # finish)
	  ~is_error:(fun _ -> c # finish)
	  ~is_aborted:(fun _ -> c # finish)
	  e;
	new Uq_engines.map_engine ~map_done:(fun _ -> `Done()) e
      method rollback () =
	if !alive then (
	  alive := false;  (* always pretend that from now on *)
	  let e =
	    new Pfs_db.async_exec c esys ~expect:[Postgresql.Command_ok] 
	      (if !prepared then
		 "ROLLBACK PREPARED '" ^ tname ^ "'"
	       else
		 "ROLLBACK") in
	  Uq_engines.when_state
	    ~is_done:(fun _ -> c # finish)
	    ~is_error:(fun _ -> c # finish)
	    ~is_aborted:(fun _ -> c # finish)
	    e;
	  new Uq_engines.map_engine ~map_done:(fun _ -> `Done()) e
	)
	else 
	  new Uq_engines.epsilon_engine (`Done ()) esys
      method return () =
	if !alive then (
	  alive := false;
	  c # finish;
	)
      method event_system = esys
    end
  )


let ro_transact_for c =
  let esys = Lazy.force !db_esys in
  let alive = ref true in
  ( object
      method connection = 
	if not !alive then 
	  failwith "Nn_db: connection is already returned to pool";
	c
      method return () =
	if !alive then (
	  alive := false;  (* always pretend that from now on *)
	  if c#idle then (
	    Netlog.logf `Debug
	      "Nn_db.ro_transact: Returning idle db conn %d" (Oo.id c);
	    Queue.push c ro_conns
	  )
	  else (
	    Netlog.logf `Debug 
	      "Nn_db.ro_transact: Dropping bad db conn %d" (Oo.id c);
	    c # finish;
	  );
	  decr ro_conns_n;
	  !ro_conns_signal()
	)
      method event_system = esys
    end
  )


let rec ro_transact() =
  let esys = Lazy.force !db_esys in
  let config = Lazy.force !config in
  let c_opt =
    if not (Queue.is_empty ro_conns) then (
      let c = Queue.take ro_conns in
      Netlog.logf `Debug "Nn_db.ro_transact: reusing db conn %d" (Oo.id c);
      Some c
    )
    else
      if !ro_conns_n < config#db_ro_connections_max then
	let c = new Pfs_db.connect config in
	Netlog.logf `Debug "Nn_db.ro_transact: new db conn %d" (Oo.id c);
	(* The default READ COMMITTED isolation level is just right for us *)
	let _e =
	  new Pfs_db.async_exec c esys ~expect:[Postgresql.Command_ok] 
	    "START TRANSACTION" in
	Some c
      else
	None in
  match c_opt with
    | Some c ->
	incr ro_conns_n;
	let t = ro_transact_for c in
	eps_e (`Done t) esys
    | None ->
	Netlog.logf `Debug "Nn_db.ro_transact: waiting";
	( match !ro_conns_e with
	    | Some e ->
		e ++ ro_transact
	    | None ->
		let (e,signal) = Uq_engines.signal_engine esys in
		ro_conns_e := Some e;
		ro_conns_signal := (fun () ->
				      ro_conns_e := None;
				      ro_conns_signal := (fun () -> ());
				      signal(`Done());
				   );
		e ++ ro_transact
	)


let with_ro_trans f =
  let esys = Lazy.force !db_esys in
  ro_transact()
  ++ (fun ro_db ->
	f ro_db#connection esys
	>> (fun st ->
	      ro_db#return();
	      st
	   )
     )


let commit_all_prepared_transactions() =
  (** There can be at most one prepared transaction for our database.
      Check whether it appears in the pg_prepared_xacts system view,
      and if so, commit it.
   *)
  let config = Lazy.force !config in
  let tname =
    "plasma_" ^ (match config#db_dbname with Some n -> n | None -> "") in
  let c = new Pfs_db.connect config in
  let r =
    c # exec
      ~expect:[Postgresql.Tuples_ok]
      ("SELECT * FROM pg_prepared_xacts WHERE gid = '" ^ tname ^ "'") in
  if r # ntuples > 0 then (
    let _r =
      c # exec 
	~expect:[Postgresql.Command_ok]
	("COMMIT PREPARED '" ^ tname ^ "'") in
    ()
  )


let datastore_list_e conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 let l =
		   List.map
		     (fun row ->
			match row with
			  | [ id_s; identity; size_s; enabled_s ] ->
			      { Nn_datastores.id = int_of_string id_s;
				identity = identity;
				size = Int64.of_string size_s;
				enabled = (enabled_s = "t");
				node = None;
			      }
			  | _ -> assert false
		     )
		     r#get_all_lst in
		 `Done l
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       "SELECT id,identity,size,enabled FROM datastore"
    )


let datastore_list_ro_e() =
  with_ro_trans datastore_list_e


let datastore_find_e ~identity conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [] ->
		       `Done None
		   | [ [ id_s; size_s; enabled_s ] ] ->
		       let ds =
			 { Nn_datastores.id = int_of_string id_s;
			   identity = identity;
			   size = Int64.of_string size_s;
			   enabled = (enabled_s = "t");
			   node = None;
			 } in
		       `Done (Some ds)
		   | _ -> assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| (* $1: *) identity |]
       "SELECT id,size,enabled FROM datastore WHERE identity = $1"
    )


let datastore_find_ro_e ~identity =
  with_ro_trans (datastore_find_e ~identity)


let blockalloc_enlarge_e ~id ~size ~old_size conn esys =
  (** Modifies the blockalloc table so that rows for [size-old_size] new
      blocks are added to it. The new blocks are marked as unallocated.
   *)
  assert(size > old_size);
  let delta = Int64.sub size old_size in
  let old_sizeH = Int64.div old_size bmaprowsizeL in
  let old_sizeL = Int64.rem old_size bmaprowsizeL in
  let sizeH = Int64.div size bmaprowsizeL in
  let sizeL = Int64.rem size bmaprowsizeL in
  
  (* First make a plan: *)
  let do_upd =
    (* whether to extend an existing row *)
    if old_sizeL > 0L then
      let blkidx = Int64.mul old_sizeH bmaprowsizeL in        (* which row *)
      let n = min (Int64.add old_sizeL delta) bmaprowsizeL in
              (* new size of blkmap *)
      Some(blkidx, Int64.to_int n)
    else
      None in
  let do_ins =
    (* list of rows to insert *)
    let first_blkidx =                 (* first row to insert *)
      if old_sizeL = 0L then 
	Int64.mul old_sizeH bmaprowsizeL
      else
	Int64.mul (Int64.succ old_sizeH) bmaprowsizeL in
    let last_blkidx =                  (* last row to insert *)
      if sizeL = 0L then
	Int64.mul (Int64.pred sizeH) bmaprowsizeL
      else
	Int64.mul sizeH bmaprowsizeL in
    let last_n =                       (* length of last row *)
      if sizeL = 0L then
	bmaprowsize
      else
	Int64.to_int sizeL in
    (first_blkidx, last_blkidx, last_n) in
  
  let exec_upd (blkidx,n) =
    new Uq_engines.seq_engine
       (new Pfs_db.async_exec conn esys 
	  ~expect:[Postgresql.Tuples_ok]
	  ~params:[| (* $1: *) string_of_int id;
	             (* $2: *) Int64.to_string blkidx
          |]
	  "SELECT blkmap FROM blockalloc WHERE datastore = $1 AND blkidx = $2"
       )
       (fun r ->
	  if r#ntuples <> 1 then
	    new Uq_engines.epsilon_engine
	      (`Error (Failure "Nn_db.blockalloc_enlarge_e: missing row"))
	      esys
	  else (
	    let blkmap = r # getvalue 0 0 in
	    let new_blkmap =
	      blkmap ^ 
		(String.make (n - String.length blkmap) '0') in
	    new Pfs_db.async_exec conn esys 
	      ~expect:[Postgresql.Command_ok]
	      ~params:[| (* $1: *) string_of_int id;
	                 (* $2: *) Int64.to_string blkidx;
                         (* $3: *) new_blkmap  (* hope no escaping req'ed *)
              |]
	      "UPDATE blockalloc SET blkmap = $3 \
               WHERE datastore = $1 AND blkidx = $2"
	  )
       )
  in

  let exec_single_ins blkidx n =
    let blkmap = String.make n '0' in
    new Uq_engines.map_engine
      ~map_done:(fun _ -> `Done())
      (new Pfs_db.async_exec conn esys 
	 ~expect:[Postgresql.Command_ok]
	 ~params:[| (* $1: *) string_of_int id;
	            (* $2: *) Int64.to_string blkidx;
                     (* $3: *) blkmap  (* hope no escaping req'ed *)
	 |]
	 "INSERT INTO blockalloc (datastore,blkidx,blkmap,blkcnt) \
          VALUES ($1, $2, $3, 0)"
      )
  in

  let rec exec_ins (first_blkidx, last_blkidx, last_n) =
    if first_blkidx < last_blkidx then
      new Uq_engines.seq_engine
	(exec_single_ins first_blkidx bmaprowsize)
	(fun _ ->
	   exec_ins (Int64.add first_blkidx bmaprowsizeL, last_blkidx, last_n)
	)
    else
      if first_blkidx = last_blkidx then
	exec_single_ins last_blkidx last_n
      else
	new Uq_engines.epsilon_engine (`Done ()) esys
  in

  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done())
    (match do_upd with
       | None ->
	   exec_ins do_ins
       | Some u ->
	   new Uq_engines.seq_engine
	     (exec_upd u)
	     (fun _ -> exec_ins do_ins)
    )
    


let datastore_set_e ~id ~identity ~size ~enabled conn esys =
   new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done())
     (let params =
	[| (* $1: *) string_of_int id;
	   (* $2: *) identity;
	   (* $3: *) Int64.to_string size;
	   (* $4: *) if enabled then "t" else "f"
	|] in
       new Pfs_db.async_exec conn esys 
	~expect:[Postgresql.Command_ok]
	~params
	 "UPDATE datastore \
          SET identity=$2, size=$3, enabled=$4 \
          WHERE id=$1"
     )


let datastore_new_e ~id ~identity ~size ~enabled conn esys =
  new Uq_engines.seq_engine
    (let params =
	[| (* $1: *) string_of_int id;
	   (* $2: *) identity;
	   (* $3: *) Int64.to_string size;
	   (* $4: *) if enabled then "t" else "f"
	|] in
       new Pfs_db.async_exec conn esys 
	~expect:[Postgresql.Command_ok]
	~params
	 "INSERT INTO datastore (id,identity,size,enabled) \
          VALUES ($1,$2,$3,$4)"
    )
    (fun _ ->
       blockalloc_enlarge_e ~id ~size ~old_size:0L conn esys
    )


let datastore_upd_e ~id ~identity ~size ~enabled conn esys =
  if size <= 0L then
    invalid_arg "Nn_db.datastore_upd_e: size is non-positive";
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Uq_engines.seq_engine
       (new Pfs_db.async_exec conn esys 
	  ~expect:[Postgresql.Tuples_ok]
	  ~params:[| (* $1: *) string_of_int id |]
	  "SELECT size FROM datastore WHERE id = $1"
       )
       (fun r ->
	  (* if no row was found by SELECT, the record is new.
             Otherwise check whether size has changed
	   *)
	  if r # ntuples = 0 then
	    (* case new row: *)
	    datastore_new_e ~id ~identity ~size ~enabled conn esys
	  else (
	    let db_size =
	      Int64.of_string (r # getvalue 0 0) in
	    if db_size > size then
	      new Uq_engines.epsilon_engine
		(`Error(Failure "Nn_db.datastore_upd_e: Cannot shrink size"))
		esys
	    else
	      if db_size < size then
		(* case enlarge datastore: also add new blockalloc rows *)
		new Uq_engines.seq_engine
		  (blockalloc_enlarge_e ~id ~size ~old_size:db_size conn esys)
		  (fun _ ->
		     (* followed by the datastore update *)
		     datastore_set_e ~id ~identity ~size ~enabled conn esys
		  )
	      else
		(* unchanged size: only overwrite the datastore row *)
		datastore_set_e ~id ~identity ~size ~enabled conn esys
	  )
       )
    )


let datastore_del_e ~id conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
      ~expect:[Postgresql.Command_ok]
      ~params:[| (* $1: *) string_of_int id |]
      "DELETE FROM datastore WHERE id = $1"
    )


let revision_get_e conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [ [ rev ] ] ->
		       `Done rev
		   | _ ->
		       assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       "SELECT revstr FROM revision"
    )


let revision_get_ro_e () =
  with_ro_trans revision_get_e



let revision_upd_e ~revstr conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) revstr |]
       "UPDATE revision SET revstr = $1"
    )


let blockalloc_list_e ~datastore conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 let l =
		   List.map
		     (fun row ->
			match row with
			  | [ blkidx_s; blkmap ] ->
			      (Int64.of_string blkidx_s, blkmap)
			  | _ -> assert false
		     )
		     r#get_all_lst in
		 `Done l
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| (* $1: *) string_of_int datastore |]
       "SELECT blkidx,blkmap FROM blockalloc WHERE datastore = $1"
    )


let blockalloc_list_ro_e ~datastore =
  with_ro_trans (blockalloc_list_e ~datastore)


let blockalloc_upd_e ~datastore ~blkidx ~blkmap conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) string_of_int datastore; 
	          (* $2: *) Int64.to_string blkidx;
	          (* $3: *) blkmap
               |]
       "UPDATE blockalloc SET blkmap = $3 WHERE datastore = $1 AND blkidx = $2"
    )


let ftype_of_char =
  function
    | 'r' -> `ftype_regular
    | 'd' -> `ftype_directory
    | 'l' -> `ftype_symlink
    | _ -> failwith "Nn_db.ftype_of_char"

let char_of_ftype =
  function
    | `ftype_regular -> 'r'
    | `ftype_directory -> 'd'
    | `ftype_symlink -> 'l'

let create_time secs nsecs =
  if nsecs < 0 || nsecs >= 1_000_000_000 then
    invalid_arg "Nn_db.create_time";
  { Pfs_rpcapi_aux.tsecs = secs; tnsecs = nsecs }


let inode_listall_e conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 `Done(List.map
			 (function
			    | [ id_s ] ->
				Int64.of_string id_s
			    | _ ->
				assert false
			 )
			 r#get_all_lst)
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| |]
       "SELECT id FROM inode"
    )


let inode_get_e ~id conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [] -> 
		       `Done None
		   | [ [ ftype_s; username; groupname; mode_s; eof_s; 
			 mtime1_s; mtime2_s; ctime1_s; ctime2_s; 
			 blimit_s; repfactor_s; field1; seqno_s; crverf_s ] ] ->
		       let ii =
			 { Pfs_rpcapi_aux.
			   filetype = ftype_of_char ftype_s.[0];
			   usergroup = {
			     Pfs_rpcapi_aux.user = username;
			     group = groupname
			   };
			   mode = int_of_string mode_s;
			   eof = Int64.of_string eof_s;
			   mtime = 
			     create_time
			       (Int64.of_string mtime1_s)
			       (int_of_string mtime2_s);
			   ctime = 
			     create_time
			       (Int64.of_string ctime1_s)
			       (int_of_string ctime2_s);
			   blocklimit = Int64.of_string blimit_s;
			   replication = int_of_string repfactor_s;
			   field1 = field1;
			   seqno = Int64.of_string seqno_s;
			   create_verifier = Int64.of_string crverf_s
			 } in
		       `Done(Some ii)
		   | _ -> assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| (* $1: *) Int64.to_string id |]
       "SELECT ftype,username,groupname,mode,eof,mtime1,mtime2,ctime1,ctime2, \
               blimit,repfactor,field1,seqno,crverf \
        FROM inode WHERE id = $1"
    )


let inode_get_ro_e ~id =
  with_ro_trans (inode_get_e ~id)


let inode_maxid_e conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [ [ max_s ] ] ->
		       if max_s = "" then
			 `Done 0L
		       else
			 `Done(Int64.of_string max_s)
		   | _ -> 
		       assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       "SELECT MAX(id) FROM inode"
    )
  
let inode_maxid_ro_e () =
 with_ro_trans inode_maxid_e


let inode_ins_e ~id ii conn esys =
  let module A = Pfs_rpcapi_aux in
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) Int64.to_string id;
	          (* $2: *) String.make 1 (char_of_ftype ii.A.filetype);
	          (* $3: *) ii.A.usergroup.A.user;
	          (* $4: *) ii.A.usergroup.A.group;
	          (* $5: *) string_of_int ii.A.mode;
    	          (* $6: *) Int64.to_string ii.A.eof;
                  (* $7: *) Int64.to_string ii.A.mtime.A.tsecs;
                  (* $8: *) string_of_int ii.A.mtime.A.tnsecs;
                  (* $9: *) Int64.to_string ii.A.ctime.A.tsecs;
                  (* $10: *) string_of_int ii.A.ctime.A.tnsecs;
	          (* $11: *) Int64.to_string ii.A.blocklimit;
	          (* $12: *) string_of_int ii.A.replication;
	          (* $13: *) ii.A.field1;
	          (* $14: *) Int64.to_string ii.A.seqno;
	          (* $15: *) Int64.to_string ii.A.create_verifier
               |]
       "INSERT INTO inode (id,ftype,username,groupname,mode,eof, \
                           mtime1,mtime2,ctime1,ctime2, \
                           blimit,repfactor,field1, seqno, crverf) \
        VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, \
                $14, $15)"
    )
    

let inode_upd_e ~id ii conn esys =
  let module A = Pfs_rpcapi_aux in
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) Int64.to_string id;
	          (* $2: *) ii.A.usergroup.A.user;
	          (* $3: *) ii.A.usergroup.A.group;
	          (* $4: *) string_of_int ii.A.mode;
    	          (* $5: *) Int64.to_string ii.A.eof;
	          (* $6: *) Int64.to_string ii.A.mtime.A.tsecs;
                  (* $7: *) string_of_int ii.A.mtime.A.tnsecs;
                  (* $8: *) Int64.to_string ii.A.ctime.A.tsecs;
                  (* $9: *) string_of_int ii.A.ctime.A.tnsecs;
	          (* $10: *) Int64.to_string ii.A.blocklimit;
	          (* $11: *) string_of_int ii.A.replication;
	          (* $12: *) ii.A.field1;
	          (* $13: *) Int64.to_string ii.A.seqno;
	          (* $14: *) Int64.to_string ii.A.create_verifier
               |]
       "UPDATE inode \
          SET username = $2, groupname = $3, mode = $4, eof = $5, \
              mtime1 = $6, mtime2 = $7, ctime1 = $8, ctime2 = $9,
              blimit = $10, repfactor = $11, field1 = $12, seqno = $13, \
              crverf = $14 \
          WHERE id = $1"
    )


let inode_upd_time_e ~id ~mtime ~ctime conn esys =
  let module A = Pfs_rpcapi_aux in
  if mtime <> None || ctime <> None then (
    new Uq_engines.map_engine
      ~map_done:(fun _ -> `Done ())
      (new Pfs_db.async_exec conn esys 
	 ~expect:[Postgresql.Command_ok]
	 (let set_mtime_clause =
	    match mtime with
	      | None -> ""
	      | Some t ->
		  sprintf "mtime1 = %Ld, mtime2 = %d" 
		    t.A.tsecs t.A.tnsecs in
	  let set_ctime_clause =
	    match ctime with
	      | None -> ""
	      | Some t ->
		  sprintf "ctime1 = %Ld, ctime2 = %d" 
		    t.A.tsecs t.A.tnsecs in
	  let set_clause =
	    set_mtime_clause ^ 
	      (if set_mtime_clause <> "" && set_ctime_clause <> "" then
		 ", " else "") ^ 
	      set_ctime_clause in
	  sprintf
	    "UPDATE inode SET %s WHERE id = %Ld"
	    set_clause id
	 )
      )
  )
  else
    eps_e (`Done ()) esys


let inode_del_e ~id conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) Int64.to_string id;
               |]
       "DELETE FROM inode WHERE id = $1"
    )


let inodeblocks_getall_e ~inode ~dsid conn esys = (* for fsck *)
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 `Done
		   (List.map
		      (fun row ->
			 match row with
			   | [ blkidx_s; block_s ] ->
			       (Int64.of_string blkidx_s,
				Int64.of_string block_s)
			   | _ ->
			       assert false
		      )
		      (r#get_all_lst)
		   )
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| (* $1: *) Int64.to_string inode;
         (* $2: *) string_of_int dsid
       |]
       "SELECT blkidx,block \
        FROM inodeblocks WHERE inode = $1 AND datastore = $2 \
        ORDER BY blkidx"
    )
    
    

let inodeblocks_get_e ~inode ~blkidx ~len conn esys =
  if len < 0L then invalid_arg "inodeblocks_get_e";
  if blkidx > Int64.sub Int64.max_int len then invalid_arg "inodeblocks_get_e";
  let ds_list = Nn_datastore_news.list ~wait:true () in
  (* TODO: If we ever allow reads on the slaves, we have to get rid of this
     list lookup
   *)
  let ds_ht = Hashtbl.create 7 in
  List.iter
    (fun ds ->
       Hashtbl.replace ds_ht ds.Nn_datastores.id ds
    )
    ds_list;
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 `Done
		   (List.flatten
		      (List.map
			 (fun row ->
			    match row with
			      | [ blkidx_s; ds_id_s; block_s ] ->
				  let ds_id = int_of_string ds_id_s in
				  ( try
				      let ds = Hashtbl.find ds_ht ds_id in
				      let node =
					match ds.Nn_datastores.node with
					  | None -> ""
					  | Some addr -> 
					      let (h,p) = 
						Plasma_util.host_port_of_sockaddr addr in
					      sprintf "%s:%d" h p in
				      [
					{ Pfs_rpcapi_aux.index = 
					    Int64.of_string blkidx_s;
					  node = node;
					  node_alive = false; (* don't know *)
					  identity = ds.Nn_datastores.identity;
					  block = Int64.of_string block_s;
					  checksum = None;
					  safetrans_id = 0L;
					  safetrans_tmo = 0L;
					  safetrans_vfy = 0L;
					  inode_seqno = 0L;
					}
				      ]
				    with
				      | Not_found ->
					  Netlog.logf `Warning
					    "Lookup for datastore ID %s failed"
					    ds_id_s;
					  []
				  )
			      | _ -> assert false
			 )
			 (r#get_all_lst)
		      )
		   )
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| (* $1: *) Int64.to_string inode;
                  (* $2: *) Int64.to_string blkidx;
                  (* $3: *) Int64.to_string (Int64.add blkidx len);
               |]
       "SELECT blkidx,datastore,block \
        FROM inodeblocks WHERE inode = $1 AND blkidx >= $2 AND blkidx < $3 \
        ORDER BY blkidx"
    )


let inodeblocks_get_ro_e ~inode ~blkidx ~len =
 with_ro_trans (inodeblocks_get_e ~inode ~blkidx ~len)



let inodeblocks_del_e ~inode ~blkidx ~len conn esys =
  if len < 0L then invalid_arg "inodeblocks_del_e";
  if blkidx > Int64.sub Int64.max_int len then invalid_arg "inodeblocks_del_e";
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
      ~expect:[Postgresql.Command_ok]
      ~params:[| (* $1: *) Int64.to_string inode;
                 (* $2: *) Int64.to_string blkidx;
                 (* $3: *) Int64.to_string (Int64.add blkidx len);
              |]
      "DELETE FROM inodeblocks \
       WHERE inode = $1 AND blkidx >= $2 AND blkidx < $3"
    )


let inodeblocks_single_ins_e ~inode bi ds conn esys =
  try
    let ds_id = ds.Nn_datastores.id in
    new Uq_engines.map_engine
      ~map_done:(fun _ -> `Done ())
      (new Pfs_db.async_exec conn esys 
	 ~expect:[Postgresql.Command_ok]
	 ~params:[| (* $1: *) Int64.to_string inode;
	            (* $2: *) Int64.to_string bi.Pfs_rpcapi_aux.index;
	            (* $3: *) string_of_int ds_id;
   	            (* $4: *) Int64.to_string bi.Pfs_rpcapi_aux.block;
         |]
	 "INSERT INTO inodeblocks (inode,blkidx,datastore,block) \
          VALUES ($1, $2, $3, $4)"
      )
  with
    | Not_found ->
	new Uq_engines.epsilon_engine
	  (`Error(Failure "inodeblocks_ins_e: ds identity not found"))
	  esys


let inodeblocks_multi_ins_e ~inode bi_list ds_list conn esys =
  try
    let ds_ht = Hashtbl.create 5 in
    List.iter
      (fun ds ->
	 Hashtbl.replace 
	   ds_ht ds.Nn_datastores.identity ds.Nn_datastores.id
      )
      ds_list;

    let copy_in_lines =
      List.map
	(fun bi ->
	   let identity = bi.Pfs_rpcapi_aux.identity in
	   let ds_id = 
	     try Hashtbl.find ds_ht identity
	     with Not_found ->
	       failwith ("identity not found: " ^ identity) in

	   sprintf "%Ld\t%Ld\t%d\t%Ld" 
	     inode 
	     bi.Pfs_rpcapi_aux.index
	     ds_id
	     bi.Pfs_rpcapi_aux.block
	)
	bi_list in
(*
List.iter
  (fun line -> Netlog.log `Debug line)
  copy_in_lines;
 *)

    new Uq_engines.map_engine
      ~map_done:(fun _ -> `Done ())
      (new Pfs_db.async_exec conn esys 
	 ~expect:[Postgresql.Command_ok]
	 ~copy_in_lines
	 "COPY inodeblocks (inode,blkidx,datastore,block) \
          FROM STDIN"
      )
  with
    | Not_found ->
	new Uq_engines.epsilon_engine
	  (`Error(Failure "inodeblocks_ins_e: ds identity not found"))
	  esys


let inodeblocks_ins_e ~inode bi_list conn esys =
  match bi_list with
    | [] ->
	new Uq_engines.epsilon_engine (`Done()) esys
    | [bi] ->
	let identity = bi.Pfs_rpcapi_aux.identity in
	datastore_find_e ~identity conn esys
	++ (function
	      | Some ds ->
		  inodeblocks_single_ins_e ~inode bi ds conn esys
	      | None ->
		  failwith ("identity not found: " ^ identity)
	   )
    | _ ->
	datastore_list_e conn esys
	++ (fun ds_list ->
	      inodeblocks_multi_ins_e ~inode bi_list ds_list conn esys
	   )


let names_get_e ~name conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [ [ inode_s ] ] ->
		       `Done(Some(Int64.of_string inode_s))
		   | [ ] ->
		       `Done None
		   | _ ->
		       assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| name |]
       "SELECT inode FROM names WHERE name = $1"
    )


let names_get_ro_e ~name =
  with_ro_trans (names_get_e ~name)


let like_special_re = Pcre.regexp "[%_\\\\]"

let like_special_subst = Pcre.subst "\\$&"

let names_list_e ~name conn esys =
  let name = if name = "/" then "" else name in
  let name_esc = 
    Pcre.replace ~rex:like_special_re ~itempl:like_special_subst name in
  let p1 = name_esc ^ "/_%" in
  let p2 = string_of_int (String.length name + 2) in
  Netlog.logf `Debug "names_list(%s): p1=%s p2=%s" name p1 p2;
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 Netlog.logf `Debug "list: ntuples=%d" r#ntuples;
		 `Done
		   (List.map
		      (fun row ->
			 match row with
			   | [ name; inode_s ] -> 
			       Netlog.logf `Debug "list: found %s" name;
			       (name, Int64.of_string inode_s)
			   | _ -> assert false
		      )
		      r#get_all_lst
		   )
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| p1; p2 |]
       "SELECT SUBSTR(name,$2),inode \
        FROM names \
        WHERE name LIKE $1
          AND STRPOS(SUBSTR(name,$2),'/') = 0"
    )

let names_list_ro_e ~name =
  with_ro_trans (names_list_e ~name)


let names_rev_get_e ~inode conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 `Done
		   (List.map
		      (fun row ->
			 match row with
			   | [ name ] -> name
			   | _ -> assert false
		      )
		      r#get_all_lst)
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| Int64.to_string inode |]
       "SELECT name FROM names WHERE inode = $1"
    )


let names_rev_get_ro_e ~inode =
 with_ro_trans (names_rev_get_e ~inode)


let names_count_e ~inode conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun r ->
		 match r#get_all_lst with
		   | [ [ count_s ] ] ->
		       `Done(int_of_string count_s)
		   | _ ->
		       assert false
	      )
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Tuples_ok]
       ~params:[| Int64.to_string inode |]
       "SELECT COUNT(*) FROM names WHERE inode = $1"
    )


let names_count_ro_e ~inode =
 with_ro_trans (names_count_e ~inode)


let names_ins_e ~name ~inode conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) name;
 	          (* $2: *) Int64.to_string inode;
               |]
       "INSERT INTO names (name, inode) \
        VALUES ($1, $2)"
    )


let names_del_e ~name conn esys =
  new Uq_engines.map_engine
    ~map_done:(fun _ -> `Done ())
    (new Pfs_db.async_exec conn esys 
       ~expect:[Postgresql.Command_ok]
       ~params:[| (* $1: *) name
               |]
       "DELETE FROM names WHERE name = $1"
    )


let exec_e m conn esys =
  match m with
    | `Datastore_upd(id,identity,size,enabled) ->
	datastore_upd_e
	  ~id ~identity ~size ~enabled conn esys
    | `Datastore_del id ->
	datastore_del_e
	  ~id  conn esys
    | `Revision_upd rev_id ->
	revision_upd_e
	  ~revstr:rev_id conn esys
    | `Blockalloc_upd(datastore,blkidx,blkmap) ->
	blockalloc_upd_e
	  ~datastore ~blkidx ~blkmap conn esys
    | `Inode_ins(id,ii) ->
	inode_ins_e ~id ii conn esys
    | `Inode_upd(id,ii) ->
	inode_upd_e ~id ii conn esys
    | `Inode_upd_time(id,mtime,ctime) ->
	inode_upd_time_e ~id ~mtime ~ctime conn esys
    | `Inode_del id ->
	inode_del_e ~id conn esys
    | `Inodeblocks_ins(inode,bi) ->
	inodeblocks_ins_e ~inode bi conn esys
    | `Inodeblocks_del(inode,blkidx,len) ->
	inodeblocks_del_e ~inode ~blkidx ~len conn esys 
    | `Names_ins(name,inode) ->
	names_ins_e ~name ~inode conn esys
    | `Names_del name ->
	names_del_e ~name conn esys
