(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: nfs3d_manager.ml 239 2010-06-23 16:49:03Z gerd $ *)

(* mount -t nfs -o ro,soft,proto=tcp,port=2801,mountport=2800,mountproto=tcp,nfsvers=3,nolock,noacl,nordirplus 127.0.0.1:/test /mnt *)

module PA = Plasma_rpcapi_aux
module NA = Nfs3_aux

open Plasma_util.Operators
open Printf

type conf =
    { clustername : string;
      namenodes : (string * int) list;
    }

type dircache =
    (int64 * int64, (string * int64) array * float) Hashtbl.t
    (* maps [(inode, cookieverf)] to the array of [(name,inode)] members, plus
       time of last access
     *)


type rundata =
    { conf : conf;
      cluster : Plasma_client.plasma_cluster;
      clients : (Unix.sockaddr, string) Hashtbl.t;
      (** Maps socket address to a secret that prooves the mount *)
      rootfh : int64;
      blocksize : int;
      blockpool : Netsys_mem.memory_pool;
      dircache : (Unix.sockaddr, dircache) Hashtbl.t;
      esys : Unixqueue.event_system;
    }

type nfs_error =
    [ `nfs3err_perm 
    | `nfs3err_noent 
    | `nfs3err_io 
    | `nfs3err_nxio 
    | `nfs3err_acces 
    | `nfs3err_exist 
    | `nfs3err_xdev 
    | `nfs3err_nodev 
    | `nfs3err_notdir 
    | `nfs3err_isdir 
    | `nfs3err_inval 
    | `nfs3err_fbig 
    | `nfs3err_nospc 
    | `nfs3err_rofs 
    | `nfs3err_mlink 
    | `nfs3err_nametoolong 
    | `nfs3err_notempty 
    | `nfs3err_dquot 
    | `nfs3err_stale 
    | `nfs3err_remote 
    | `nfs3err_badhandle 
    | `nfs3err_not_sync 
    | `nfs3err_bad_cookie 
    | `nfs3err_notsupp 
    | `nfs3err_toosmall 
    | `nfs3err_serverfault 
    | `nfs3err_badtype 
    | `nfs3err_jukebox 
    ]


type 'a nfs_error_arg =
    [ `nfs3err_perm of 'a
    | `nfs3err_noent of 'a
    | `nfs3err_io of 'a
    | `nfs3err_nxio of 'a
    | `nfs3err_acces of 'a
    | `nfs3err_exist of 'a
    | `nfs3err_xdev of 'a
    | `nfs3err_nodev of 'a
    | `nfs3err_notdir of 'a
    | `nfs3err_isdir of 'a
    | `nfs3err_inval of 'a
    | `nfs3err_fbig of 'a
    | `nfs3err_nospc of 'a
    | `nfs3err_rofs of 'a
    | `nfs3err_mlink of 'a
    | `nfs3err_nametoolong of 'a 
    | `nfs3err_notempty of 'a
    | `nfs3err_dquot of 'a
    | `nfs3err_stale of 'a
    | `nfs3err_remote of 'a
    | `nfs3err_badhandle of 'a
    | `nfs3err_not_sync of 'a
    | `nfs3err_bad_cookie of 'a
    | `nfs3err_notsupp of 'a
    | `nfs3err_toosmall of 'a
    | `nfs3err_serverfault of 'a
    | `nfs3err_badtype of 'a
    | `nfs3err_jukebox of 'a
    ]

exception Nfs_error of nfs_error
      
exception Retry

let add_arg nfs_error arg =
  match nfs_error with
    | `nfs3err_perm -> `nfs3err_perm arg 
    | `nfs3err_noent -> `nfs3err_noent arg 
    | `nfs3err_io -> `nfs3err_io arg 
    | `nfs3err_nxio -> `nfs3err_nxio arg 
    | `nfs3err_acces -> `nfs3err_acces arg 
    | `nfs3err_exist -> `nfs3err_exist arg 
    | `nfs3err_xdev -> `nfs3err_xdev arg 
    | `nfs3err_nodev -> `nfs3err_nodev arg 
    | `nfs3err_notdir -> `nfs3err_notdir arg 
    | `nfs3err_isdir -> `nfs3err_isdir arg 
    | `nfs3err_inval -> `nfs3err_inval arg 
    | `nfs3err_fbig -> `nfs3err_fbig arg 
    | `nfs3err_nospc -> `nfs3err_nospc arg 
    | `nfs3err_rofs -> `nfs3err_rofs arg 
    | `nfs3err_mlink -> `nfs3err_mlink arg 
    | `nfs3err_nametoolong -> `nfs3err_nametoolong arg 
    | `nfs3err_notempty -> `nfs3err_notempty arg 
    | `nfs3err_dquot -> `nfs3err_dquot arg 
    | `nfs3err_stale -> `nfs3err_stale arg 
    | `nfs3err_remote -> `nfs3err_remote arg 
    | `nfs3err_badhandle -> `nfs3err_badhandle arg 
    | `nfs3err_not_sync -> `nfs3err_not_sync arg
    | `nfs3err_bad_cookie -> `nfs3err_bad_cookie arg
    | `nfs3err_notsupp -> `nfs3err_notsupp arg 
    | `nfs3err_toosmall -> `nfs3err_toosmall arg 
    | `nfs3err_serverfault -> `nfs3err_serverfault arg 
    | `nfs3err_badtype -> `nfs3err_badtype arg 
    | `nfs3err_jukebox -> `nfs3err_jukebox arg 

      
let read_multiple = 2


let new_rundata conf =
  let sesys = Unixqueue.create_unix_event_system() in
  let scluster = 
    Plasma_client.open_cluster conf.clustername conf.namenodes sesys in
  let st =
    Plasma_client.start scluster in
  let rootfh = 
    Plasma_client.lookup st "/" in
  let blocksize =
    Plasma_client.blocksize scluster in
  Plasma_client.commit st;
  Plasma_client.close_cluster scluster;

  let esys = (Netplex_cenv.self_cont()) # event_system in
  let cluster = 
    Plasma_client.open_cluster conf.clustername conf.namenodes esys in

  Plasma_client.configure_buffer cluster 64;
  (* TODO: make this user-configurable *)

  { conf = conf;
    cluster = cluster;
    clients = Hashtbl.create 7;
    rootfh = rootfh;
    blocksize = blocksize;
    blockpool = Netsys_mem.create_pool (read_multiple*blocksize);
    dircache = Hashtbl.create 7;
    esys = esys
  }

let global_rundata = ref None

let get_rundata conf =
  match !global_rundata with
    | Some rd ->
	if rd.conf == conf then
	  rd
	else
	  failwith "More than one configuration"
    | None ->
	let rd = new_rundata conf in
	global_rundata := Some rd;
	rd

let peer sess =
  let saddr = Rpc_server.get_peer_name sess in
  match saddr with
    | Unix.ADDR_INET(ip,_) ->
	Unix.ADDR_INET(ip,0)
    | _ ->
	saddr


let wrap_fh secret inode =
  secret ^ Rtypes.int8_as_string (Rtypes.int8_of_int64 inode)

let unwrap_fh fh =
  if String.length fh <> 16 then
    failwith "unwrap_fh";
  let secret = String.sub fh 0 8 in
  (secret, Rtypes.int64_of_int8 (Rtypes.read_int8 fh 8))

let wrap_sess_fh sess rd inode =
  let saddr = peer sess in
  try
    let secret = Hashtbl.find rd.clients saddr in
    wrap_fh secret inode
  with
    | Not_found ->
	raise (Nfs_error `nfs3err_serverfault)
	
	  
let unwrap_sess_fh sess rd fh =
  let saddr = peer sess in
  if String.length fh <> 16 then
    raise (Nfs_error `nfs3err_badhandle);
  try
    let exp_secret = Hashtbl.find rd.clients saddr in
    let (secret, inode) = unwrap_fh fh in
    if secret <> exp_secret then
      raise (Nfs_error `nfs3err_badhandle);
    inode
  with
    | Not_found ->
	raise (Nfs_error `nfs3err_serverfault)

let to_uid user =
  (* TODO *)
  try Int64.of_string user
  with _ -> 65534L   (* "nobody" *)

let to_gid group =
  (* TODO *)
  try Int64.of_string group
  with _ -> 65534L   (* "nogroup" *)
  
let to_nfstime t =
  { NA.seconds = t.PA.tsecs;
      nseconds = Int64.of_int t.PA.tnsecs
  }


let to_fattr3 inode ii =
  { NA.fa_type = (match ii.PA.filetype with
		    | `ftype_regular -> `nf3reg
		    | `ftype_directory -> `nf3dir
		    | `ftype_symlink -> `nf3lnk
		 );
    fa_mode = Int64.of_int (ii.PA.mode land 0o777);
    fa_nlink = 1L;    (* FIXME *)
    fa_uid = to_uid ii.PA.usergroup.PA.user;
    fa_gid = to_gid ii.PA.usergroup.PA.group;
    fa_size = ii.PA.eof;
    fa_used = 0L;    (* FIXME *)
    fa_rdev = { NA.specdata1 = 0L; specdata2 = 0L };
    fa_fsid = 0L;
    fa_fileid = inode;
    fa_atime = to_nfstime ii.PA.mtime;  (* no atime *)
    fa_mtime = to_nfstime ii.PA.mtime;
    fa_ctime = to_nfstime ii.PA.ctime;
  }

let to_nfserr 
      ?(efhier = `nfs3err_perm)
      ?(retry_on_conflict = false)
      (pe : Plasma_client.errno) : nfs_error =
  match pe with
    | `enotrans -> `nfs3err_serverfault
    | `efailedcommit -> `nfs3err_serverfault
    | `elongtrans -> `nfs3err_serverfault
    | `efailed -> `nfs3err_serverfault
    | `eperm -> `nfs3err_perm
    | `enoent -> `nfs3err_noent
    | `eaccess -> `nfs3err_acces
    | `eexist -> `nfs3err_exist
    | `efhier -> efhier
    | `einval -> `nfs3err_inval
    | `efbig -> `nfs3err_fbig
    | `enospc -> `nfs3err_nospc
    | `erofs -> `nfs3err_rofs
    | `enametoolong -> `nfs3err_nametoolong
    | `econflict -> 
	if retry_on_conflict then raise Retry else `nfs3err_serverfault
    | `ecoord -> `nfs3err_serverfault
    | `enonode -> `nfs3err_serverfault
    | `etbusy -> `nfs3err_serverfault
    | `enoient -> `nfs3err_stale
    | `eio -> `nfs3err_io


let to_nfserr_arg
      ?efhier
      ?retry_on_conflict
      pe arg : _ nfs_error_arg =
  add_arg (to_nfserr ?efhier ?retry_on_conflict pe) arg

(* NFS *)

let pause =
  [| 0.01; 0.02; 0.04; 0.08; 0.16; 0.32; 0.64; 1.0;
     2.0; 4.0; 8.0; 16.0; 32.0; 60.0
  |]

let transaction_e rd name f =
  (** Creates a transaction [t] and runs [f t esys]. *)
  (* TODO: avoid serialization *)
  let result = ref `Aborted in
  let rec next_attempt retry =
    Netlog.logf `Debug "Starting transaction for %s" name;
    Plasma_client.start_e rd.cluster
    ++ (fun t ->
	  let e = f t rd.esys in
	  Plasma_util.failsafe_e e 
	  ++ (fun st ->
		let success, do_retry =
		  match st with
		    | `Done _ -> true, false
		    | `Error Retry -> false, true
		    | _ -> false, false in
		( if success then (
		    Netlog.logf `Debug
		      "Committing transaction for %s" name;
		    Plasma_client.commit_e t
		  )
		  else (
		    Netlog.logf `Debug "Aborting transaction for %s" name;
		    Plasma_client.abort_e t
		  )
		)
		++ (fun () ->
		      if do_retry then (
			let d =
			  if retry < Array.length pause then
			    pause.(retry)
			  else
			    pause.(Array.length pause - 1) in
			Netlog.logf `Debug
			  "Retrying transaction for %s after %f seconds"
			  name d;
			Plasma_util.delay_engine d
			  (fun _ -> next_attempt (retry+1))
			  rd.esys
		      )
		      else (
			result := st;
			eps_e (`Done ()) rd.esys
		      )
		   )
	     )
       ) in
  next_attempt 0
    >> (function
	  | `Done () -> `Done !result
	  | `Error e -> `Error e
	  | `Aborted -> `Aborted
       )
      

(* --- read calls  --- *)

let proc_lookup conf sess name_in_dir emit =
  let rd = get_rundata conf in
  try
    let dir_inode = unwrap_sess_fh sess rd name_in_dir.NA.dirop_dir in
    Plasma_util.rpc_srv_reply "lookup" sess emit
      (transaction_e rd "lookup"
	 (fun t esys ->
	    Plasma_client.dir_lookup_e
	      t
	      dir_inode
	      name_in_dir.NA.dirop_name
	    ++ (fun file_inode ->
		  Plasma_client.get_inodeinfo_e 
		    t
		    file_inode
		  ++ (fun file_ii ->
			eps_e (`Done(file_inode,file_ii)) esys
		     )
	       )
	    >> (function
		  | `Done (file_inode,file_ii) ->
		      let file_fh = wrap_sess_fh sess rd file_inode in
		      let file_attr = to_fattr3 file_inode file_ii in
		      let ok =
			{ NA.lookup_object = file_fh;
			  lookup_obj_attributes = Some file_attr;
			  lookup_dir_attributes = None
			} in
		      `Done(`nfs3_ok ok)
		  | `Error (Plasma_client.Plasma_error plasma_err) ->
		      `Done(to_nfserr_arg
			      ~efhier:`nfs3err_notdir
			      ~retry_on_conflict:true
			      plasma_err
			      { NA.lookup_dir_attributes_fail = None }
			    :> NA.lookup3res)
		  | `Error err ->
		      `Error err
		  | `Aborted ->
		      `Aborted
	       )
	 )
      )
  with
    | Nfs_error e ->
	let r =
	  (add_arg e { NA.lookup_dir_attributes_fail = None } :> NA.lookup3res)
	in
	emit r


let get_ii conf sess fh proc_name f_ok f_err f_nfserr emit =
  let rd = get_rundata conf in
  try
    let inode = unwrap_sess_fh sess rd fh in
    Plasma_util.rpc_srv_reply proc_name sess emit
      (transaction_e rd proc_name
	 (fun t esys ->
	    Plasma_client.get_inodeinfo_e
	      t
	      inode
	    ++ (fun ii ->
		  try
		    let ok = f_ok inode ii in
		    eps_e (`Done(`nfs3_ok ok)) rd.esys
		  with
		    | error -> eps_e (`Error error) rd.esys
	       )
	    >> (function
		  | `Done r -> `Done r
		  | `Error (Plasma_client.Plasma_error plasma_err) ->
		      `Done(f_err plasma_err)
		  | `Error err ->
		      `Error err
		  | `Aborted ->
		      `Aborted
	       )
	 )
      )
  with
    | Nfs_error e ->
	let r = f_nfserr e in
	emit r



let proc_getattr conf sess fh emit =
  get_ii
    conf sess fh "getattr"
    (fun inode ii ->
       let attr = to_fattr3 inode ii in
       { NA.getattr_obj_attributes = attr }
    )
    (fun plasma_err ->
       (to_nfserr
	  ~retry_on_conflict:true
	  plasma_err
	:> NA.getattr3res)
    )
    (fun nfs_err -> (nfs_err :> NA.getattr3res))
    emit


let proc_access conf sess (fh, perms) emit = 
  (* fake for now - we always grant everything *)
  get_ii
    conf sess fh "access"
    (fun inode ii ->
       let attr = to_fattr3 inode ii in
       { NA.access_obj_attributes_ok = Some attr;
	 access_rights_ok = Int64.logand perms 0x3fL;
       }
    )
    (fun plasma_err ->
       (to_nfserr_arg
	  ~retry_on_conflict:true
	  plasma_err
	  { NA.access_obj_attributes_fail = None }
	:> NA.access3res)
    )
    (fun nfs_err ->
       (add_arg nfs_err { NA.access_obj_attributes_fail = None }
	:> NA.access3res))
    emit


let proc_readlink conf sess fh emit =
  get_ii
    conf sess fh "readlink"
    (fun inode ii ->
       let attr = to_fattr3 inode ii in
       match ii.PA.filetype with
	 | `ftype_symlink ->
	     { NA.readlink_symlink_attributes_ok = Some attr;
	       readlink_data = ii.PA.field1
	     }
	 | _ ->
	     raise(Plasma_client.Plasma_error `einval)
	     (* TODO: can also return attr here:
	       { NA.readlink_symlink_attributes_fail = Some attr }
	      *)
    )
    (fun plasma_err ->
       (to_nfserr_arg
	  ~retry_on_conflict:true
	  plasma_err
	  { NA.readlink_symlink_attributes_fail = None }
	:> NA.readlink3res)
    )
    (fun nfs_err ->
       (add_arg nfs_err { NA.readlink_symlink_attributes_fail = None }
	:> NA.readlink3res))
    emit


let proc_fsstat conf sess fh emit =
  (* fake so far *)
  let rd = get_rundata conf in
  try
    let inode = unwrap_sess_fh sess rd fh in
    if inode <> rd.rootfh then
      raise(Nfs_error `nfs3err_stale);
    let r =
      { NA.fsstat_obj_attributes_ok = None;
	fsstat_tbytes = 0L;
	fsstat_fbytes = 0L;
	fsstat_abytes = 0L;
	fsstat_tfiles = 0L;
	fsstat_ffiles = 0L;
	fsstat_afiles = 0L;
	fsstat_invarsec = 0L;
      } in
    emit (`nfs3_ok r)
  with
    | Nfs_error e ->
	emit
	  (add_arg e { NA.fsstat_obj_attributes_fail = None } :> NA.fsstat3res)


let proc_fsinfo conf sess fh emit = 
  (* fake so far *)
  let rd = get_rundata conf in
  try
    let inode = unwrap_sess_fh sess rd fh in
    if inode <> rd.rootfh then
      raise(Nfs_error `nfs3err_stale);
    let r =
      { NA.fsinfo_obj_attributes_ok = None;
	fsinfo_rtmax = Int64.of_int (read_multiple * rd.blocksize);
	fsinfo_rtpref = Int64.of_int rd.blocksize;
	fsinfo_rtmult = 512L;  (* nobody knows what this means *)
	fsinfo_wtmax = Int64.of_int rd.blocksize;
	fsinfo_wtpref = Int64.of_int rd.blocksize;
	fsinfo_wtmult = 512L;  (* nobody knows what this means *)
	fsinfo_dtpref = 32768L;
	fsinfo_maxfilesize = Int64.max_int;
	fsinfo_time_delta = { NA.seconds = 0L; nseconds = 1L };
	fsinfo_properties =
	  Int64.logor
	    (Int64.logor
	       (Int64.logor
		  (Rtypes.int64_of_uint4 NA.fsf3_link)
		  (Rtypes.int64_of_uint4 NA.fsf3_symlink))
	       (Rtypes.int64_of_uint4 NA.fsf3_homogeneous))
	    (Rtypes.int64_of_uint4 NA.fsf3_cansettime);
      } in
    emit (`nfs3_ok r)
  with
    | Nfs_error e ->
	emit
	  (add_arg e { NA.fsinfo_obj_attributes_fail = None } :> NA.fsinfo3res)



let proc_pathconf conf sess fh emit =
  get_ii
    conf sess fh "pathconf"
    (fun inode ii ->
       let attr = to_fattr3 inode ii in
       { NA.pathconf_obj_attributes_ok = Some attr;
	 pathconf_linkmax = 32767L;   (* actually unlimited *)
	 pathconf_name_max = 32767L;  (* actually unlimited *)
	 pathconf_no_trunc = true;
	 pathconf_chown_restricted = false;
	 pathconf_case_insensitive = false;
	 pathconf_case_preserving = true;
       }
    )
    (fun plasma_err ->
       (to_nfserr_arg
	  ~retry_on_conflict:true
	  plasma_err
	  { NA.pathconf_obj_attributes_fail = None }
	:> NA.pathconf3res)
    )
    (fun nfs_err ->
       (add_arg nfs_err { NA.pathconf_obj_attributes_fail = None }
	:> NA.pathconf3res))
    emit

(*
let memory_string rd size =
  let mlen = Netsys_mem.init_string_bytelen size in
  let mem = 
    if mlen <= Netsys_mem.pool_block_size rd.blockpool then
      Netsys_mem.pool_alloc_memory rd.blockpool
    else
      Bigarray.Array1.create Bigarray.char Bigarray.c_layout mlen in
  let voffs,_ = Netsys_mem.init_string mem 0 size in
  let s = (Netsys_mem.as_value mem voffs : string) in
  let m = Bigarray.Array1.sub mem voffs size in
  (* mem: the underlying bigarray
     m:   the bigarray which is physically identical with s
     s:   the string inside mem
   *)
  (mem, m, s)
 *)


let prefix_mstring (m:Xdr_mstring.mstring) p =
  assert(p <= m#length);
  ( object
      method length = p
      method blit_to_string = m#blit_to_string
      method blit_to_memory = m#blit_to_memory
      method as_string = m#as_string
      method as_memory = m#as_memory
      method preferred = m#preferred
    end
  )


let proc_read conf sess (fh, offset, count) emit =
  let rd = get_rundata conf in
  try
    let inode = unwrap_sess_fh sess rd fh in
    let blocksizeL = Int64.of_int rd.blocksize in

    (* offset and count are specified as unsigned int64 in the NFS
       protocol. Here, they are passed as signed int64, i.e. negative
       values are large ones.

       If offset is negative, we immediately return 0 and set the
       EOF flag.

       If count is negative, we replace it with max_int.
     *)
    if offset < 0L then
      emit
	(`nfs3_ok
	   { NA.read_file_attributes_ok = None;
	     read_count_ok = 0L;
	     read_eof = true;
	     read_data = (Xdr_mstring.memory_based_mstrings #
			    create_from_string "" 0 0 false
			 )
	   })
    else (
      let index = Int64.div offset blocksizeL in
      let bpos = Int64.sub offset (Int64.mul index blocksizeL) in

      let max_count = 
	if bpos = 0L then
	  read_multiple * rd.blocksize
	else
	  rd.blocksize in
      let max_countL = Int64.of_int max_count in
      let count =
	if count < 0L then 
	  max_count
	else
	  Int64.to_int (min count max_countL) in

      let full_name =
	sprintf "read(%Ld,%Ld,%d)" inode offset count in

      let m = 
	if count <= Netsys_mem.pool_block_size rd.blockpool then
	  Netsys_mem.pool_alloc_memory rd.blockpool
	else
	  Bigarray.Array1.create Bigarray.char Bigarray.c_layout count in

      let plan = ref [] in
      let left = ref count in
      let p = ref 0 in
      while !left > 0 do
	let l = min !left rd.blocksize in
	plan := (!p, l) :: !plan;
	left := !left - l;
	p := !p + l;
      done;
      if !plan = [] then
	plan := [0,0];

      (* TODO: We can get ECONFLICT here. Handle this *)

      (* TODO: The multi-read is hairy. Better restart reading when any
	 problem is detected (e.g. change of ii)
       *)

      Plasma_util.rpc_srv_reply full_name sess emit
	(Uq_engines.meta_engine (*  transaction_e rd full_name *)
	   ( (Uq_engines.msync_engine
		(List.map
		   (fun (p,l) ->
		      Plasma_client.read_e
			rd.cluster
			inode
			(Int64.add offset (Int64.of_int p))
			(`Memory m)
			p
			l
		      ++ (fun (n,eof,ii) ->
			    if n < l then
			      Bigarray.Array1.fill
				(Bigarray.Array1.sub m n (l-n))
				'\000';
			    let n' =
			      if n > 0 then p+n else 0 in
			    eps_e (`Done (n',eof,ii)) rd.esys
			 )
		   )
		   !plan
		)
		(fun (n,eof,ii) (acc_n,acc_eof,acc_ii_opt) ->
		   if n > acc_n then
		     (n,eof,Some ii)
		   else
		     if acc_ii_opt = None then
		       (acc_n,acc_eof,Some ii)
		     else
		       (acc_n,acc_eof,acc_ii_opt)
		)
		(0,false,None)
		rd.esys
	     )
	     ++ (fun (n, eof, ii_opt) ->
		   assert(n <= count);
		   let ii =
		     match ii_opt with None -> assert false|Some ii -> ii in
		   let r =
		     { NA.read_file_attributes_ok = Some(to_fattr3 inode ii);
		       read_count_ok = Int64.of_int n;
		       read_eof = eof;
		       read_data = (Xdr_mstring.memory_based_mstrings #
				      create_from_memory m 0 n false
				   )
		     } in
		   eps_e (`Done(`nfs3_ok r)) rd.esys
		   )
	     >> (function
		   | `Done r -> `Done r
		   | `Error (Plasma_client.Plasma_error plasma_err) ->
		       `Done(to_nfserr_arg
			       ~retry_on_conflict:true
			       plasma_err
			       { NA.read_file_attributes_fail = None }
			     :> NA.read3res)
		   | `Error err ->
		       `Error err
		   | `Aborted ->
		       `Aborted
		)
	   )
	)
    )
  with
    | Nfs_error e ->
	emit
	  (add_arg e { NA.read_file_attributes_fail = None } :> NA.read3res)


let null_verf = 0L

let readdir_from_cache rd saddr inode verf =
  try
    if verf = null_verf then raise Not_found;
    let dircache = Hashtbl.find rd.dircache saddr in
    let (entries, _) = Hashtbl.find dircache (inode, verf) in
    let now = Unix.gettimeofday() in
    Hashtbl.replace dircache (inode, verf) (entries, now);
    Some entries
  with
    | Not_found ->
	None

let readdir_e rd t saddr inode =
  (** not cached, so we have to query the list *)
  Plasma_client.list_inode_e
    t
    inode
  ++ (fun entries ->
	let entries = Array.of_list entries in
	let dircache =
	  try Hashtbl.find rd.dircache saddr
	  with
	    | Not_found ->
		let dircache = Hashtbl.create 13 in
		Hashtbl.add rd.dircache saddr dircache;
		dircache in
	let new_verf = ref (Plasma_rng.random_int64()) in
	while !new_verf = null_verf || 
	    Hashtbl.mem dircache (inode,!new_verf) 
	do
	  new_verf := Int64.succ !new_verf
	done;
	let now = Unix.gettimeofday() in
	Hashtbl.add dircache (inode,!new_verf) (entries, now);
	eps_e (`Done (entries, !new_verf)) rd.esys
     )

let clean_dircache rd saddr =
  (* limit the number of entries stored there *)
  (* TODO: call this function periodically *)
  let n_keep = 4 in
  try
    let dircache = Hashtbl.find rd.dircache saddr in
    let keep = ref [] in
    let k = ref 0 in
    let del = ref [] in
    let now = Unix.gettimeofday() in
    Hashtbl.iter
      (fun (inode,verf) (_,t) ->
	 if now -. t < 60.0 then (
	   keep := (inode, verf, t) :: !keep;
	   incr k;
	   if !k > n_keep then (  
	     (** remove oldest entry from [keep] *)
	     let (o_inode, o_verf, o_t) =
	       List.fold_left
		 (fun (o_inode, o_verf, o_t) (inode, verf, t) ->
		    if t < o_t then
		      (inode, verf, t)
		    else
		      (o_inode, o_verf, o_t)
		 )
		 (List.hd !keep)
		 (List.tl !keep) in
	     keep := 
	       List.filter
		 (fun (inode, verf, _) -> inode <> o_inode || verf <> o_verf)
		 !keep;
	     del := (o_inode, o_verf) :: !del;
	     decr k
	   )
	 )
	 else
	   del := (inode, verf) :: !del
      )
      dircache;
    List.iter
      (fun (inode, verf) ->
	 Hashtbl.remove dircache (inode,verf)
      )
      !del
  with
    | Not_found -> ()
  

let readdir_response count entries k_start verf =
  let count =
    if count < 0L then
      max_int
    else
      Int64.to_int (min (Int64.of_int max_int) count) in

  let nfs_entry k (name,inode) =
    let ne =
      { NA.entry_fileid = inode;
	entry_name = name;
	entry_cookie = Int64.of_int (k+1);
	entry_nextentry = None
      } in
    let name_size =
      (String.length name - 1) / 4 + 5 in
    let ne_size =
      name_size + 20 (* 20: remaining stuff in entry3 *) in
    (ne, ne_size) in

  let rec nfs_entries k xdr_size =
    if k >= Array.length entries then
      (None, true)
    else (
      let (ne, ne_size) = nfs_entry k entries.(k) in
      if count - ne_size >= xdr_size then (
	let (l', eof) = nfs_entries (k+1) (xdr_size + ne_size) in
	(Some { ne with NA.entry_nextentry = l' }, eof)
      )
      else
	(None, false)
    ) in

  let dirlist3_size = 8 in
  let resok_size = 12 in

  let l, eof = nfs_entries k_start (dirlist3_size + resok_size) in
  
  { NA.readdir_dir_attributes_ok = None;
    readdir_cookieverf_ok = verf;
    readdir_reply = 
      { NA.dl_entries = l;
	dl_eof = eof
      }
  }

let proc_readdir_common conf sess (fh, cookie, verf_s) 
                        f_ok emit fail coerce =
  let rd = get_rundata conf in
  let saddr = peer sess in
  let verf = Rtypes.int64_of_int8 (Rtypes.read_int8 verf_s 0) in
  try
    let inode = unwrap_sess_fh sess rd fh in
    (** First check whether we can respond from dircache: *)
    match readdir_from_cache rd saddr inode verf with
      | Some entries ->
	  (** Found in the cache. The [cookie] is simply interpreted as the
	      index in [names] where to continue
	   *)
	  if cookie < 0L || cookie >= Int64.of_int max_int then
	    raise(Nfs_error `nfs3err_bad_cookie);
	  let cookie = Int64.to_int cookie in
	  if cookie > Array.length entries then
	    raise(Nfs_error `nfs3err_bad_cookie);
	  clean_dircache rd saddr;
	  let r = f_ok entries cookie verf_s in
	  emit (`nfs3_ok r)

      | None ->
	  if verf <> 0L then
	    raise (Nfs_error `nfs3err_bad_cookie);

	  (** not in the cache: start transaction, and read directory *)
	  
	  Plasma_util.rpc_srv_reply "readdir" sess emit
	    (transaction_e rd "readdir"
	       (fun t esys ->
		  readdir_e rd t saddr inode
		    ++ (fun (entries, new_verf) ->
			  let new_verf_s =
			    Rtypes.int8_as_string
			      (Rtypes.int8_of_int64 new_verf) in
			  let r =
			    f_ok entries 0 new_verf_s in
			  let rr = 
			    r.NA.readdir_reply in
			  if rr.NA.dl_entries = None && 
			     not rr.NA.dl_eof
			  then
			    eps_e (`Done(`nfs3err_toosmall fail)) rd.esys
			  else
			    eps_e (`Done(`nfs3_ok r)) rd.esys
		       )
		  >> (function
			| `Done r -> `Done r
			| `Error (Plasma_client.Plasma_error plasma_err) ->
			    `Done(coerce
				    (to_nfserr_arg
				       ~retry_on_conflict:true
				       plasma_err
				       fail))
			| `Error err ->
			    `Error err
			| `Aborted ->
			    `Aborted
		     )
	       )
	    )
  with
   | Nfs_error e ->
	emit
	  (coerce (add_arg e fail))

let proc_readdir conf sess (fh, cookie, verf_s, count) emit =
  let fail =
    { NA.readdir_dir_attributes_fail = None } in
  let f =
    readdir_response count in
  let coerce x =
    ( x :> NA.readdir3res ) in
  proc_readdir_common conf sess (fh, cookie, verf_s) f emit fail coerce


let proc_readdirplus conf sess (fh, cookie, verf, dircount, maxcount) emit = 
  (** In the RFC [dircount] is unspecified. Obviously, it refers to
      some buffer size in the Solaris implementation. We simply return
      NOTSUPP so clients fall back to readdir.
   *)
  let fail =
    { NA.readdirp_dir_attributes_fail = None } in
  emit (`nfs3err_notsupp fail)



(* --- write calls  --- *)

(* Return ROFS for now: *)

let proc_setattr conf sess (fh, sattr, guard) emit =
  let fail =
    { NA.setattr_obj_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_write conf sess (fh, offset, count, stable, data) emit = 
  let fail =
    { NA.write_file_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)
  
let proc_create conf sess (name_in_dir, how) emit =
  let fail =
    { NA.create_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_mkdir conf sess (name_in_dir, sattr) emit =
  let fail =
    { NA.mkdir_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_symlink conf sess (name_in_dir, data) emit =
  let fail =
    { NA.symlink_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_remove conf sess name_in_dir emit =
  let fail =
    { NA.remove_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_rmdir conf sess name_in_dir emit =
  let fail =
    { NA.rmdir_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_rename conf sess (name1_in_dir,name2_in_dir) emit =
  let fail =
    { NA.rename_fromdir_wcc_fail = 
	{ NA.before = None; after = None };
      NA.rename_todir_wcc_fail = 
	{ NA.before = None; after = None };
    } in
  emit (`nfs3err_rofs fail)

let proc_link conf sess (fh, name_in_dir) emit =
  let fail =
    { NA.link_file_attributes_fail = None;
      NA.linkdir_wcc_fail = 
	{ NA.before = None; after = None };
    } in
  emit (`nfs3err_rofs fail)


let proc_commit conf sess (fh, offset, count) emit =
  let fail =
    { NA.commit_file_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let proc_mknod conf sess (name_in_dir, data) emit = 
  let fail =
    { NA.mknod_dir_wcc_fail = 
	{ NA.before = None; after = None } } in
  emit (`nfs3err_rofs fail)

let setup_nfs srv conf =
  Nfs3_srv.NFS.V3.bind_async
    ~proc_null:(fun sess () emit -> emit())
    ~proc_lookup:(proc_lookup conf)
    ~proc_getattr:(proc_getattr conf)
    ~proc_access:(proc_access conf)
    ~proc_readlink:(proc_readlink conf)
    ~proc_read:(proc_read conf)
    ~proc_readdir:(proc_readdir conf)
    ~proc_readdirplus:(proc_readdirplus conf)
    ~proc_fsstat:(proc_fsstat conf)
    ~proc_fsinfo:(proc_fsinfo conf)
    ~proc_pathconf:(proc_pathconf conf)
    ~proc_setattr:(proc_setattr conf)
    ~proc_write:(proc_write conf)
    ~proc_create:(proc_create conf)
    ~proc_mkdir:(proc_mkdir conf)
    ~proc_symlink:(proc_symlink conf)
    ~proc_remove:(proc_remove conf)
    ~proc_rmdir:(proc_rmdir conf)
    ~proc_rename:(proc_rename conf)
    ~proc_link:(proc_link conf)
    ~proc_commit:(proc_commit conf)
    ~proc_mknod:(proc_mknod conf)
    srv;
  let m_none =
    Rpc_server.auth_none in
  let m_sys = 
    Rpc_auth_sys.server_auth_method 
      ~lookup_hostname:false
      ~require_privileged_port:false
      () in
  Rpc_server.set_auth_methods srv [m_none;m_sys]


(* MOUNT *)

let proc_mnt conf sess path emit =
  let rd = get_rundata conf in
  let saddr = peer sess in
  let r =
    if path <> "/" ^ conf.clustername then
      `mnt3err_noent
    else (
      let secret =
	try
	  Hashtbl.find rd.clients saddr
	with
	  | Not_found ->
	      let secret = Plasma_rng.random_bytes 8 in
	      Hashtbl.add rd.clients saddr secret;
	      secret in
      let fh = wrap_fh secret rd.rootfh in
      `mnt3_ok
	{ NA.fhandle = fh;
	  auth_flavors = [| 0 |]
	}
    ) in
  emit r
      

let name_of_saddr saddr =
  match saddr with
    | Unix.ADDR_INET(ip,_) ->
	Unix.string_of_inet_addr ip
    | Unix.ADDR_UNIX _ ->
	"localhost"

let proc_dump conf sess  () emit =
  let rd = get_rundata conf in
  let l =
    Hashtbl.fold
      (fun saddr _ acc ->
	 Some
	   { NA.ml_hostname = name_of_saddr saddr;
	     ml_directory = "/" ^ conf.clustername;
	     ml_next = acc
	   }
      )
      rd.clients
      None in
  emit l


let proc_umnt conf sess path emit =
  let rd = get_rundata conf in
  if path = "/" ^ conf.clustername then (
    let saddr = peer sess in
    Hashtbl.remove rd.clients saddr;
    Hashtbl.remove rd.dircache saddr;
  );
  emit ()


let proc_umntall conf sess () emit =
  let rd = get_rundata conf in
  let saddr = peer sess in
  Hashtbl.remove rd.clients saddr;
  Hashtbl.remove rd.dircache saddr;
  emit ()


let proc_export conf sess () emit =
  emit
    (Some
       { NA.ex_dir = "/" ^ conf.clustername;
	 ex_groups = None;
	 ex_next = None
       })


let setup_mount srv conf =
  Nfs3_srv.MOUNT.V3.bind_async
    ~proc_null:(fun _ _ emit -> emit())
    ~proc_mnt:(proc_mnt conf)
    ~proc_dump:(proc_dump conf)
    ~proc_umnt:(proc_umnt conf)
    ~proc_umntall:(proc_umntall conf)
    ~proc_export:(proc_export conf)
    srv;
  let m_none =
    Rpc_server.auth_none in
  let m_sys = 
    Rpc_auth_sys.server_auth_method 
      ~lookup_hostname:false
      ~require_privileged_port:false
      () in
  Rpc_server.set_auth_methods srv [m_none;m_sys]


let parse_global_conf cf =
  let namenodes =
    cf#resolve_section cf#root_addr "namenodes" in
  match namenodes with
    | [] ->
        failwith "missing 'namenodes' config"
    | _ :: _ :: _ ->
        failwith "more than one 'namenodes' config"
    | [nn] ->
	let addrs = Plasma_util.node_list cf nn in
        if addrs = [] then
          failwith "nfs config: no namenodes";
	let nnlist =
          List.map
            (fun addr ->
               try Plasma_util.parse_host_port addr
               with _ ->
		 failwith ("bad 'addr' parameter in 'namenodes.node': " ^ addr)
            )
            addrs in
	let clustername =
            try
              let p = cf#resolve_parameter nn "clustername" in
              cf#string_param p
            with
              | Not_found ->
                  failwith "missing 'namenodes.clustername' parameter" in
	{ clustername = clustername;
	  namenodes = nnlist
	}


let global_conf = ref None

let get_global_conf cf =
  match !global_conf with
    | None ->
	let c = parse_global_conf cf in
	global_conf := Some c;
	c
    | Some c ->
	c


let nfs3_factory() =
  Rpc_netplex.rpc_factory
    ~configure:(fun cf cfaddr ->
		  get_global_conf cf
	       )
    ~hooks:(fun conf ->
	      ( object
		  inherit Netplex_kit.empty_processor_hooks()
		end
	      )
	   )
    ~setup:(fun srv conf ->
	      setup_nfs srv conf
	   )
    ~name:"nfs3"
    ()

let mount3_factory() =
  Rpc_netplex.rpc_factory
    ~configure:(fun cf cfaddr ->
		  get_global_conf cf
	       )
    ~hooks:(fun conf ->
	      ( object
		  inherit Netplex_kit.empty_processor_hooks()
		end
	      )
	   )
    ~setup:(fun srv conf ->
	      setup_mount srv conf
	   )
    ~name:"mount3"
    ()

let factory () =
  new Netplex_kit.protocol_switch_factory 
    "nfs"
    [ "nfs3", nfs3_factory();
      "mount3", mount3_factory()
    ]
