(*
  Copyright 2010 Gerd Stolpmann

  This file is part of Plasma, a distributed filesystem and a
  map/reduce computation framework. Unless you have a written license
  agreement with the copyright holder (Gerd Stolpmann), the following
  terms apply:

  Plasma is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  Plasma is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.

*)
(* $Id: plasma_shm.ml 239 2010-06-23 16:49:03Z gerd $ *)

type shm_id = [`Shm_id of int ]

type shm_condition = [ `Normal | `Tight ]

class type shm_manager =
object
  method new_id : unit -> shm_id
  method book_allocation : shm_id -> int -> unit
  method shm_condition : shm_condition
end


class null_shm_manager () : shm_manager =
object(self)
  val mutable next_id = 0
  method new_id() =
    let id = next_id in
    next_id <- id+1;
    `Shm_id id
  method book_allocation _ _ = ()
  method shm_condition = `Normal
end


(** For the Netplex shm manager, we keep track of allocated memory
    per component. For each component there is a variable with name
    "plasma_shm.sum.<cid>" where <cid> is a component ID. This variable
    contains the sum of all allocated shm.

    We configure that Netplex deletes this variable automatically when
    the component terminates.

    There is also "plasma_shm.directory" which is a list of all
    of the above sum variables. The mutex "plasma_shm.dir_mutex" is used
    to control access to the directory variable.
 *)

module Shm_sum_var =
  Netplex_sharedvar.Make_var_type (struct type t = int64 end)

module Shm_dir_var =
  Netplex_sharedvar.Make_var_type (struct type t = string list end)

module Shm_watermarks_var =
  Netplex_sharedvar.Make_var_type (struct type t = int64 * int64 end)

module Shm_condition_var =
  Netplex_sharedvar.Make_var_type (struct type t = shm_condition end)


let dir_name = "plasma_shm.directory" 
let wm_name = "plasma_shm.watermarks"
let cond_name = "plasma_shm.condition"

let announce_sum_var sum_name =
  let mutex_name = "plasma_shm.dir_mutex" in
  let created = Netplex_sharedvar.create_var ~enc:true dir_name in
  if created then (
    Shm_dir_var.set dir_name [sum_name]
  )
  else (
    ignore(Netplex_sharedvar.wait_for_enc_value dir_name);
    let mutex = Netplex_mutex.access mutex_name in
    Netplex_mutex.lock mutex;
    let l = Shm_dir_var.get dir_name in
    Shm_dir_var.set dir_name (sum_name :: l);
    Netplex_mutex.unlock mutex;
  )

let get_global_sum() =
  let sum_names = 
    try Shm_dir_var.get dir_name 
    with Netplex_sharedvar.Sharedvar_not_found _ -> [] in
  List.fold_left
    (fun acc sum_name ->
       try
	 Int64.add acc (Shm_sum_var.get sum_name) 
       with
	 | Netplex_sharedvar.Sharedvar_not_found _ -> acc
    )
    0L
    sum_names


let get_watermarks() =
  try
    Shm_watermarks_var.get wm_name
  with
    | Netplex_sharedvar.Sharedvar_not_found _ -> (Int64.max_int, Int64.max_int)


let init_watermarks low high =
  let _ =
     Netplex_sharedvar.create_var ~enc:true wm_name in
  Shm_watermarks_var.set wm_name (low,high)


let get_condition() =
  try
    Shm_condition_var.get cond_name
  with
    | Netplex_sharedvar.Sharedvar_not_found _ -> `Normal
  

let set_condition c =
  let _ =
     Netplex_sharedvar.create_var ~enc:true cond_name in
  Shm_condition_var.set cond_name c


let push_sum sum =
  let cont = Netplex_cenv.self_cont() in
  let sum_name = 
    string_of_int (Oo.id cont#container_id) in
  let created =  Netplex_sharedvar.create_var ~own:true ~enc:true sum_name in
  Shm_sum_var.set sum_name sum;
  if created then
    announce_sum_var sum_name;
  let gsum = get_global_sum() in
  Netlog.logf `Debug
    "Plasma_shm: local_sum = %Ld global_sum = %Ld"
    sum gsum;
  let (low,high) = get_watermarks() in
  let c = get_condition() in
  if c = `Normal && gsum >= high then
    set_condition `Tight
  else
    if c = `Tight && gsum <= low then
      set_condition `Normal


class netplex_shm_manager() : shm_manager =
object(self)

  val mutable next_id = 0
  method new_id() =
    let id = next_id in
    next_id <- id+1;
    `Shm_id id

  val account = Hashtbl.create 5

  method book_allocation (`Shm_id id) size =
    Hashtbl.replace account id size;
    let sum =
      Hashtbl.fold
	(fun _ size acc -> Int64.add acc (Int64.of_int size))
	account
	0L in
    push_sum sum

  method shm_condition =
    let c = get_condition() in
    Netlog.logf `Debug
      "Plasma_shm: condition = %s"
      ( match c with
	  | `Normal -> "normal"
	  | `Tight -> "tight"
      );
    c

end


module Netplex_shm_manager = struct
  let post_add_hook ctrl =
    ctrl # add_plugin Netplex_sharedvar.plugin;
    ctrl # add_plugin Netplex_mutex.plugin

  let init ~low ~high =
    init_watermarks low high
end
